/* ipuz-clues.c
 *
 * Copyright 2024 Jonathan Blandford <jrb@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * SPDX-License-Identifier: (LGPL-2.1-or-later OR MIT)
 */


#include "libipuz-config.h"
#include <glib/gi18n-lib.h>

#include "ipuz-clue-sets.h"
#include "ipuz-clues.h"
#include "ipuz-private.h"


static guint             ipuz_clues_real_get_n_clue_sets        (IpuzClues                *self);
static IpuzClueDirection ipuz_clues_real_clue_set_get_dir       (IpuzClues                *self,
                                                                 guint                     index);
const gchar             *ipuz_clues_real_clue_set_get_label     (IpuzClues                *self,
                                                                 IpuzClueDirection         direction);
static GArray           *ipuz_clues_real_get_clues              (IpuzClues                *self,
                                                                 IpuzClueDirection         direction);
static void              ipuz_clues_real_foreach_clue           (IpuzClues                *self,
                                                                 IpuzCluesForeachClueFunc  func,
                                                                 gpointer                  user_data);
static guint             ipuz_clues_real_get_n_clues            (IpuzClues                *self,
                                                                 IpuzClueDirection         direction);
static IpuzClue         *ipuz_clues_real_get_clue_by_id         (IpuzClues                *self,
                                                                 IpuzClueId               *clue_id);
static void              ipuz_clues_real_unlink_clue            (IpuzClues                *self,
                                                                 IpuzClue                 *clue);
static gboolean          ipuz_clues_real_get_id_by_clue         (IpuzClues                *self,
                                                                 const IpuzClue           *clue,
                                                                 IpuzClueId               *clue_id);
static gchar            *ipuz_clues_real_get_clue_string_by_id  (IpuzClues                *self,
                                                                 IpuzClueId               *clue_id);
static gchar            *ipuz_clues_real_get_guess_string_by_id (IpuzClues                *self,
                                                                 IpuzClueId               *clue_id);
static gboolean          ipuz_clues_real_clue_guessed           (IpuzClues                *self,
                                                                 IpuzClue                 *clue,
                                                                 gboolean                 *correct);
static IpuzClue         *ipuz_clues_real_find_clue_by_number    (IpuzClues                *self,
                                                                 IpuzClueDirection         direction,
                                                                 gint                      number);
static IpuzClue         *ipuz_clues_real_find_clue_by_label     (IpuzClues                *self,
                                                                 IpuzClueDirection         direction,
                                                                 const gchar              *label);
static IpuzClue         *ipuz_clues_real_find_clue_by_coord     (IpuzClues                *self,
                                                                 IpuzClueDirection         direction,
                                                                 const IpuzCellCoord      *coord);


G_DEFINE_INTERFACE (IpuzClues, ipuz_clues, IPUZ_TYPE_PUZZLE);


static void
ipuz_clues_default_init (IpuzCluesInterface *iface)
{
  iface->get_n_clue_sets = ipuz_clues_real_get_n_clue_sets;
  iface->clue_set_get_dir = ipuz_clues_real_clue_set_get_dir;
  iface->clue_set_get_label = ipuz_clues_real_clue_set_get_label;
  iface->get_clues = ipuz_clues_real_get_clues;
  iface->foreach_clue = ipuz_clues_real_foreach_clue;
  iface->get_n_clues = ipuz_clues_real_get_n_clues;
  iface->get_clue_by_id = ipuz_clues_real_get_clue_by_id;
  iface->unlink_clue = ipuz_clues_real_unlink_clue;
  iface->get_id_by_clue = ipuz_clues_real_get_id_by_clue;
  iface->get_clue_string_by_id = ipuz_clues_real_get_clue_string_by_id;
  iface->get_guess_string_by_id = ipuz_clues_real_get_guess_string_by_id;
  iface->clue_guessed = ipuz_clues_real_clue_guessed;
  iface->find_clue_by_number = ipuz_clues_real_find_clue_by_number;
  iface->find_clue_by_label = ipuz_clues_real_find_clue_by_label;
  iface->find_clue_by_coord = ipuz_clues_real_find_clue_by_coord;
}

/* Default implementations */
static guint
ipuz_clues_real_get_n_clue_sets (IpuzClues *self)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return 0;

  return ipuz_clue_sets_get_n_clue_sets (clue_sets);
}

static IpuzClueDirection
ipuz_clues_real_clue_set_get_dir (IpuzClues *self,
                                  guint      index)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return IPUZ_CLUE_DIRECTION_NONE;

  return ipuz_clue_sets_get_direction (clue_sets, index);
}

const gchar *
ipuz_clues_real_clue_set_get_label (IpuzClues         *self,
                                    IpuzClueDirection  direction)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return NULL;

  return ipuz_clue_sets_get_label (clue_sets, direction);
}

static GArray *
ipuz_clues_real_get_clues (IpuzClues         *self,
                           IpuzClueDirection  direction)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return NULL;

  return ipuz_clue_sets_get_clues (clue_sets, direction);
}

static void
ipuz_clues_real_foreach_clue (IpuzClues                *self,
                              IpuzCluesForeachClueFunc  func,
                              gpointer                  user_data)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return;

  for (guint i = 0; i < ipuz_clue_sets_get_n_clue_sets (clue_sets); i++)
    {
      IpuzClueDirection direction;
      GArray *clues;

      direction = ipuz_clue_sets_get_direction (clue_sets, i);
      clues = ipuz_clue_sets_get_clues (clue_sets, direction);

      for (guint j = 0; j < clues->len; j++)
        {
          IpuzClue *clue;
          IpuzClueId clue_id = {
            .direction = direction,
            .index = j,
          };

          clue = g_array_index (clues, IpuzClue *, j);
          (func) (self, direction, clue, &clue_id, user_data);
        }
    }
}

static guint
ipuz_clues_real_get_n_clues (IpuzClues         *self,
                             IpuzClueDirection  direction)
{
  GArray *clues;

  clues = ipuz_clues_get_clues (self, direction);
  if (clues)
    return clues->len;

  return 0;
}

static IpuzClue *
ipuz_clues_real_get_clue_by_id (IpuzClues  *self,
                                IpuzClueId *clue_id)
{
  GArray *clues;

  clues = ipuz_clues_get_clues (IPUZ_CLUES (self), clue_id->direction);

  if (clues && clue_id->index < clues->len)
    return g_array_index (clues, IpuzClue *, clue_id->index);

  return NULL;
}

static void
unlink_clue_from_cells (IpuzGrid *self,
                        IpuzClue *clue)
{
  for (guint i = 0; i < ipuz_clue_get_n_coords (clue); i++)
    {
      IpuzCellCoord coord;
      IpuzCell *cell = NULL;

      ipuz_clue_get_coord (clue, i, &coord);
      cell = ipuz_grid_get_cell (self, &coord);

      if (cell)
        ipuz_cell_clear_clue_direction (cell, ipuz_clue_get_direction (clue));
    }
}

static void
ipuz_clues_real_unlink_clue (IpuzClues *self,
                             IpuzClue  *clue)
{
  IpuzClueSets *clue_sets = _ipuz_puzzle_get_clue_sets (IPUZ_PUZZLE (self));

  if (clue_sets == NULL)
    return;

  g_return_if_fail (IPUZ_IS_GRID (self));

  //  g_warning ("this function won't work correctly until we add refcounting to clues\n");
  unlink_clue_from_cells (IPUZ_GRID (self), clue);
  ipuz_clue_sets_remove_clue (clue_sets, ipuz_clue_get_direction (clue), clue, TRUE);
}

static gboolean
ipuz_clues_real_get_id_by_clue (IpuzClues      *self,
                                const IpuzClue *clue,
                                IpuzClueId     *clue_id)
{
  GArray *clues;

  clue_id->direction = IPUZ_CLUE_DIRECTION_NONE;
  clue_id->index = 0;

  if (clue == NULL)
    return FALSE;

  clues = ipuz_clues_get_clues (self, ipuz_clue_get_direction (clue));

  if (clues == NULL)
    return FALSE;

  for (guint i = 0; i < clues->len; i++)
    {
      IpuzClue *tmp_clue = g_array_index (clues, IpuzClue *, i);
      if (ipuz_clue_equal (clue, tmp_clue))
        {
          clue_id->direction = ipuz_clue_get_direction (clue);
          clue_id->index = i;
          return TRUE;
        }
    }

  return FALSE;
}


static gchar *
get_string_from_clue (IpuzGrid   *self,
                      IpuzClueId *clue_id,
                      gboolean    guesses)
{
  IpuzClue *clue;
  GString *string;
  guint i;

  clue = ipuz_clues_get_clue_by_id (IPUZ_CLUES (self), clue_id);

  if (clue == NULL)
    return NULL;

  string = g_string_new (NULL);
  for (i = 0; i < ipuz_clue_get_n_coords (clue); i++)
    {
      IpuzCellCoord coord;
      const char *solution = NULL;

      ipuz_clue_get_coord (clue, i, &coord);
      if (guesses)
        {
          solution = ipuz_guesses_get_guess (ipuz_grid_get_guesses (self),
                                             &coord);
        }
      else
        {
          IpuzCell *cell;

          cell = ipuz_grid_get_cell (self, &coord);
          solution = ipuz_cell_get_solution (cell);
        }

      if (solution && solution[0] !='\0')
        g_string_append (string, solution);
      else
        g_string_append (string, "?");
    }

  return g_string_free (string, FALSE);
}

static gchar *
ipuz_clues_real_get_clue_string_by_id (IpuzClues  *self,
                                       IpuzClueId *clue_id)
{
  return get_string_from_clue (IPUZ_GRID (self), clue_id, FALSE);
}

static gchar *
ipuz_clues_real_get_guess_string_by_id (IpuzClues  *self,
                                        IpuzClueId *clue_id)
{
  return get_string_from_clue (IPUZ_GRID (self), clue_id, TRUE);
}

static gboolean
ipuz_clues_real_clue_guessed (IpuzClues *self,
                              IpuzClue  *clue,
                              gboolean  *correct)
{
  guint i;
  gboolean guessed = TRUE;
  IpuzGuesses *guesses;

  /* This implementation is grid-specific */
  g_return_val_if_fail (IPUZ_IS_GRID (self), FALSE);

  guesses = ipuz_grid_get_guesses (IPUZ_GRID (self));

  if (ipuz_clue_get_n_coords (clue) == 0)
    return FALSE;

  if (guesses == NULL)
    return FALSE;

  if (correct)
    *correct = TRUE;

  for (i = 0; i < ipuz_clue_get_n_coords (clue); i++)
    {
      IpuzCellCoord coord;
      IpuzCell *cell;

      ipuz_clue_get_coord (clue, i, &coord);
      cell = ipuz_grid_get_cell (IPUZ_GRID (self), &coord);

      if (ipuz_cell_get_initial_val (cell))
        continue;

      if (correct)
        *correct = ipuz_grid_check_cell (IPUZ_GRID (self), cell, &coord,
                                         guesses, IPUZ_GRID_CHECK_GUESS_CORRECT)
          && *correct;

      if (! ipuz_grid_check_cell (IPUZ_GRID (self), cell, &coord,
                                  guesses, IPUZ_GRID_CHECK_GUESS_MADE))
        {
          guessed = FALSE;
          break;
        }
    }

  return guessed;
}

static IpuzClue *
ipuz_clues_real_find_clue_by_number (IpuzClues         *self,
                                     IpuzClueDirection  direction,
                                     gint               number)
{
  GArray *clues = NULL;

  clues = ipuz_clues_get_clues (self, direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IpuzClue *clue;
      clue = g_array_index (clues, IpuzClue *, i);
      if (ipuz_clue_get_number (clue) == number)
        return clue;
    }

  return NULL;
}

static IpuzClue *
ipuz_clues_real_find_clue_by_label (IpuzClues         *self,
                                    IpuzClueDirection  direction,
                                    const gchar       *label)
{
  GArray *clues = NULL;

  clues = ipuz_clues_get_clues (self, direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IpuzClue *clue;
      clue = g_array_index (clues, IpuzClue *, i);
      if (g_strcmp0 (label, ipuz_clue_get_label (clue)) == 0)
        return clue;
    }
  return NULL;
}

static IpuzClue *
ipuz_clues_real_find_clue_by_coord (IpuzClues           *self,
                                    IpuzClueDirection    direction,
                                    const IpuzCellCoord *coord)
{
 GArray *clues = NULL;

  clues = ipuz_clues_get_clues (self, direction);
  if (clues == NULL)
    return NULL;

  for (guint i = 0; i < clues->len; i ++)
    {
      IpuzClue *clue;
      clue = g_array_index (clues, IpuzClue *, i);
      if (ipuz_clue_contains_coord (clue, coord))
        return clue;
    }
  return NULL;
}

/* public functions */

guint
ipuz_clues_get_n_clue_sets (IpuzClues *self)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), 0);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_n_clue_sets);

  return IPUZ_CLUES_GET_IFACE (self)->get_n_clue_sets (self);
}

IpuzClueDirection
ipuz_clues_clue_set_get_dir (IpuzClues *self,
                             guint      index)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), IPUZ_CLUE_DIRECTION_NONE);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->clue_set_get_dir);

  return IPUZ_CLUES_GET_IFACE (self)->clue_set_get_dir (self, index);
}

const gchar *
ipuz_clues_clue_set_get_label (IpuzClues         *self,
                               IpuzClueDirection  direction)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), IPUZ_CLUE_DIRECTION_NONE);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->clue_set_get_label);

  return IPUZ_CLUES_GET_IFACE (self)->clue_set_get_label (self, direction);
}

/**
 * ipuz_clues_get_clues:
 * @direction: The direction of clues to search
 *
 * Returns an array of all clues in the direction of @direction.
 *
 * Returns: (element-type IpuzClue) (transfer none): an array of all clues in the direction of @direction
 **/
GArray *
ipuz_clues_get_clues (IpuzClues         *self,
                      IpuzClueDirection  direction)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_clues);

  return IPUZ_CLUES_GET_IFACE (self)->get_clues (self, direction);
}

/**
 * ipuz_clues_foreach_clue:
 * @func: (scope call): The function to call for each clue
 * @user_data: User data to pass to @func
 *
 * Calls @func for each [struct@Ipuz.Clue] in @self.
 **/
void
ipuz_clues_foreach_clue (IpuzClues                *self,
                         IpuzCluesForeachClueFunc  func,
                         gpointer                  user_data)
{
  g_return_if_fail (IPUZ_IS_CLUES (self));

  g_assert (IPUZ_CLUES_GET_IFACE (self)->foreach_clue);

  return IPUZ_CLUES_GET_IFACE (self)->foreach_clue (self, func, user_data);
}

guint
ipuz_clues_get_n_clues (IpuzClues         *self,
                        IpuzClueDirection  direction)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), 0);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_n_clues);

  return IPUZ_CLUES_GET_IFACE (self)->get_n_clues (self, direction);
}

/**
 * ipuz_clues_get_clue_by_id:
 * @clue_id: 
 *
 * 
 *
 * Returns: (transfer none): A clue
 **/
IpuzClue *
ipuz_clues_get_clue_by_id (IpuzClues  *self,
                           IpuzClueId *clue_id)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);
  g_return_val_if_fail (clue_id != NULL, NULL);
  g_return_val_if_fail (! IPUZ_CLUE_ID_IS_UNSET (clue_id), NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_clue_by_id);

  return IPUZ_CLUES_GET_IFACE (self)->get_clue_by_id (self, clue_id);
}

/* FIXME: This should be renamed to remove clue. It was confusing to figure out what it did. :-) */
void
ipuz_clues_unlink_clue (IpuzClues *self,
                        IpuzClue  *clue)
{
  g_return_if_fail (IPUZ_IS_CLUES (self));
  g_return_if_fail (clue != NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->unlink_clue);

  return IPUZ_CLUES_GET_IFACE (self)->unlink_clue (self, clue);
}

/**
 * ipuz_clues_get_id_by_clue:
 * @clue: (nullable): The [struct@Ipuz.Clue] to search for
 * @clue_id: (out): An [struct@Ipuz.ClueId] to fill in with the id of @clue
 *
 * Finds the [struct@Ipuz.ClueId] of @clue within @self.
 *
 * If @clue doesn't exist in @self, then %FALSE is returned and
 * @clue_id's direction is set to [enum@Ipuz.ClueDirection.NONE].
 *
 * Returns: %TRUE, if @clue_id was set to the id of @clue
 **/
gboolean
ipuz_clues_get_id_by_clue (IpuzClues      *self,
                           const IpuzClue *clue,
                           IpuzClueId     *clue_id)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), FALSE);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_id_by_clue);

  return IPUZ_CLUES_GET_IFACE (self)->get_id_by_clue (self, clue, clue_id);
}

/**
 * ipuz_clues_get_clue_string_by_id:
 * @clue_id: An [struct@Ipuz.ClueId] to get the clue string for
 *
 * Returns a string containing the solution of the puzzle for a given
 * clue. This string will have '?' characters embedded within it if
 * there are cells without solutions set yet.
 *
 * Returns: (nullable) (transfer full): The solution for the puzzle
 * for @clue_id. Returns %NULL if @clue_id points to an invalid clue
 **/
gchar *
ipuz_clues_get_clue_string_by_id  (IpuzClues  *self,
                                   IpuzClueId *clue_id)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);
  g_return_val_if_fail (clue_id != NULL, NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_clue_string_by_id);

  return IPUZ_CLUES_GET_IFACE (self)->get_clue_string_by_id (self, clue_id);
}

/**
 * ipuz_clues_get_guess_string_by_id:
 * @clue_id: An [struct@Ipuz.ClueId] to get the guesses string for
 *
 * Returns a string containing the guess in the puzzle for a given
 * clue. This string will have '?' characters embedded within it if
 * there are cells not completely filled out.
 *
 * Returns: (nullable) (transfer full): The guesses for the puzzle for
 * @clue_id. Returns %NULL if @clue_id points to an invalid clue
 **/
gchar *
ipuz_clues_get_guess_string_by_id (IpuzClues  *self,
                                   IpuzClueId *clue_id)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);
  g_return_val_if_fail (clue_id != NULL, NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->get_guess_string_by_id);

  return IPUZ_CLUES_GET_IFACE (self)->get_guess_string_by_id (self, clue_id);
}

/**
 * ipuz_clues_clue_guessed:
 * @clue::
 * @correct: (out):
 *
 * 
 *
 * Returns: %TRUE, if the clue has guesses for every cell
 **/
gboolean
ipuz_clues_clue_guessed (IpuzClues *self,
                         IpuzClue  *clue,
                         gboolean  *correct)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), FALSE);
  g_return_val_if_fail (clue != NULL, FALSE);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->clue_guessed);

  return IPUZ_CLUES_GET_IFACE (self)->clue_guessed (self, clue, correct);
}

IpuzClue *
ipuz_clues_find_clue_by_number    (IpuzClues         *self,
                                   IpuzClueDirection  direction,
                                   gint               number)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->find_clue_by_number);

  return IPUZ_CLUES_GET_IFACE (self)->find_clue_by_number (self, direction, number);
}

IpuzClue *
ipuz_clues_find_clue_by_label (IpuzClues         *self,
                               IpuzClueDirection  direction,
                               const char        *label)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);
  g_return_val_if_fail (label != NULL, NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->find_clue_by_label);

  return IPUZ_CLUES_GET_IFACE (self)->find_clue_by_label (self, direction, label);
}

IpuzClue *
ipuz_clues_find_clue_by_coord (IpuzClues           *self,
                               IpuzClueDirection    direction,
                               const IpuzCellCoord *coord)
{
  g_return_val_if_fail (IPUZ_IS_CLUES (self), NULL);

  g_assert (IPUZ_CLUES_GET_IFACE (self)->find_clue_by_coord);

  return IPUZ_CLUES_GET_IFACE (self)->find_clue_by_coord (self, direction, coord);
}
